{
Piping data from and to processes and process signaling

Copyright (C) 1998-99 Free Software Foundation, Inc.

Author: Frank Heckenbach <frank@pascal.gnu.de>

This Library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public License as
published by the Free Software Foundation, version 2.

This Library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this Library; see the file COPYING.LIB.  If
not, write to the Free Software Foundation, Inc., 675 Mass Ave,
Cambridge, MA 02139, USA.
}

unit Pipe;

interface

uses GPC;

type TPStrings (Count: Cardinal) = array [1 .. Count] of ^String; (*@@two unit bug with definition in gpc.pas in programs using these two units*)

type
  TWaitPIDResult = (PIDNothing, PIDExited, PIDSignaled, PIDStopped, PIDUnknown);

  PPipeProcess = ^TPipeProcess;
  TPipeProcess = record
    PID       : Integer;         { Process ID of process forked }
    OpenPipes : Integer;         { Number of pipes to/from the process, for internal use }
    Signal    : Integer;         { Send this signal (if not 0) to the process after all pipes have been closed after some time }
    Seconds   : Integer;         { Wait so many seconds before sending the signal if the process has not terminated by itself }
    Wait      : Boolean;         { Wait for the process, even longer than Seconds seconds, after sending the signal (if any) }
    Result    : ^TWaitPIDResult; { Default nil. If a pointer to a variable is stored here, the result (cf. WaitPID) is stored here when waiting after closing the pipes }
    Status    : ^Integer;        { Default nil. If a pointer to a variable is stored here, the status (cf. WaitPID) is stored here when waiting after closing the pipes }
  end;

const
  EPipe = 600; { cannot create pipe to `%s' }
  AnyChild = - 1; { possible value for PID in WaitPID }

var
  { Default values for TPipeProcess records created by Pipe }
  DefaultPipeSignal  : Integer = 0;
  DefaultPipeSeconds : Integer = 0;
  DefaultPipeWait    : Boolean = True;

  { Signals. The variables are initialized with the signal numbers, and
    are 0 for signals not defined. }
  { POSIX signals }
  SigHUP    : asmname '_p_SIGHUP'    Integer;
  SigInt    : asmname '_p_SIGINT'    Integer;
  SigQuit   : asmname '_p_SIGQUIT'   Integer;
  SigIll    : asmname '_p_SIGILL'    Integer;
  SigAbrt   : asmname '_p_SIGABRT'   Integer;
  SigFPE    : asmname '_p_SIGFPE'    Integer;
  SigKill   : asmname '_p_SIGKILL'   Integer;
  SigSegV   : asmname '_p_SIGSEGV'   Integer;
  SigPipe   : asmname '_p_SIGPIPE'   Integer;
  SigAlrm   : asmname '_p_SIGALRM'   Integer;
  SigTerm   : asmname '_p_SIGTERM'   Integer;
  SigUsr1   : asmname '_p_SIGUSR1'   Integer;
  SigUsr2   : asmname '_p_SIGUSR2'   Integer;
  SigChld   : asmname '_p_SIGCHLD'   Integer;
  SigCont   : asmname '_p_SIGCONT'   Integer;
  SigStop   : asmname '_p_SIGSTOP'   Integer;
  SigTStp   : asmname '_p_SIGTSTP'   Integer;
  SigTTIn   : asmname '_p_SIGTTIN'   Integer;
  SigTTOu   : asmname '_p_SIGTTOU'   Integer;

  { Non-POSIX signals }
  SigTrap   : asmname '_p_SIGTRAP'   Integer;
  SigIOT    : asmname '_p_SIGIOT'    Integer;
  SigEMT    : asmname '_p_SIGEMT'    Integer;
  SigBus    : asmname '_p_SIGBUS'    Integer;
  SigSys    : asmname '_p_SIGSYS'    Integer;
  SigStkFlt : asmname '_p_SIGSTKFLT' Integer;
  SigUrg    : asmname '_p_SIGURG'    Integer;
  SigIO     : asmname '_p_SIGIO'     Integer;
  SigPoll   : asmname '_p_SIGPOLL'   Integer;
  SigCld    : asmname '_p_SIGCLD'    Integer;
  SigXCPU   : asmname '_p_SIGXCPU'   Integer;
  SigXFSz   : asmname '_p_SIGXFSZ'   Integer;
  SigVTAlrm : asmname '_p_SIGVTALRM' Integer;
  SigProf   : asmname '_p_SIGPROF'   Integer;
  SigPwr    : asmname '_p_SIGPWR'    Integer;
  SigInfo   : asmname '_p_SIGINFO'   Integer;
  SigLost   : asmname '_p_SIGLOST'   Integer;
  SigWinCh  : asmname '_p_SIGWINCH'  Integer;
  SigUnused : asmname '_p_SIGUNUSED' Integer;

{ Returns a description for a signal }
function StrSignal (Signal : Integer) : TString;

{ Sends a signal to a process. Returns True if successful. If Signal is 0,
  it doesn't send a signal, but still checks whether it would be possible
  to send a signal to the given process. }
function Kill (PID, Signal : Integer) : Boolean; asmname '_p_kill';

{ Waits for a child process to terminate or be stopped. Returns the PID of
  the process. Result will contain the information whether it terminated
  by itself, or was terminated or stopped by a signal. Status will contain
  the exit status if it terminated by itself, or the number of the signal
  otherwise. If nothing happened, and Block is False, the function will
  return 0, Result will be PIDNothing and Status will be 0. If an error
  occurred (especially on single tasking systems where WaitPID is not
  possible), the function will return - 1. Result and Status can be null,
  in which case they're ignored. }
function WaitPID (PID : Integer; var Result : TWaitPIDResult; var Status : Integer; Block : Boolean) : Integer; asmname '_p_waitpid';

{ The procedure Pipe starts a process whose name is given by ProcessName,
  with the given Parameters (can be null if no parameters) and Environment,
  and create pipes from and/or to the process' standard input/output/error.
  ProcessName is searched for in the PATH with FSearchExecutable. Any of
  ToInputFile, FromOutputFile and FromStdErrFile can be null if the
  corresponding pipe is not wanted. FromOutputFile and FromStdErrFile may
  be identical, in which case standard output and standard error are
  redirected to the same pipe. The behaviour of other pairs of files being
  identical is undefined, and useless, anyway. The files are Assigned and
  Reset or Rewritten as appropriate. Errors are shown in IOResult, @@however,
  IOResult is not checked automatically after this function, even in `(*$I+*)'
  mode. If Process is not null, a pointer to a record is stored there, from
  which the PID of the process created can be read, and by writing to which
  the action after all pipes have been closed can be changed. (The record
  is automatically Dispose'd of after all pipes have been closed.) If
  automatic waiting is turned off, the caller should get the PID from the
  record before it's Dispose'd of, and wait for the process sometime in order
  to avoid zombies. If no redirections are performed (i.e., all 3 files are
  null), the caller should wait for the process with WaitPipeProcess.
  When an error occurs, Process is not assigned to, and the state of the
  files is undefined, so be sure to check IOResult before going on.

  The procedure even works under Dos, but of course, in a limited sense:
  if ToInputFile is used, the process will not actually be started until
  ToInputFile is closed. Signal, Seconds and Wait of TPipeProcess are
  ignored, and PID does not contain a Process ID, but an internal value
  without any meaning to the caller. Result will always be PIDExited.
  So, Status is the only interesting field (but Result should also be
  checked).

  So, the only portable way to use "pipes" in both directions is to call
  `Pipe', write all the Input data to ToInputFile, close ToInputFile, and
  then read the Output and StdErr data from FromOutputFile and
  FromStdErrFile. However, since the capacity of pipes is limited, one
  should also check for Data from FromOutputFile and FromStdErrFile (using
  CanRead, IOSelect or IOSelectRead) while writing the Input data (under
  Dos, there simply won't be any data then, but checking for data doesn't
  do any harm). Please see pipedemo.pas for an example. }
(*@@IO critical*) procedure Pipe (var ToInputFile, FromOutputFile, FromStdErrFile : AnyFile; (*@@fjf265 const*) ProcessName : String; protected var Parameters : TPStrings; Environment : PCStrings; var Process : PPipeProcess);

{ Waits for a process created by Pipe as determined in the Process record.
  (Process is Dispose'd of afterwards.) Returns True if successful. }
function WaitPipeProcess (Process : PPipeProcess) : Boolean;

implementation

{$B-,I-}

{$L pipec.c}
procedure PipeInit; asmname '_p_pipeinit';
function CStrSignal (Signal : Integer) : CString; asmname 'strsignal';

function StrSignal (Signal : Integer) : TString;
begin
  StrSignal := CString2String (CStrSignal (Signal))
end;

{$ifndef MSDOS}

type
  PCFilePtr = ^CFilePtr;

  PFileProcess = ^TFileProcess;
  TFileProcess = record
    FilePtr : PAnyFile;
    OldCloseProc : TCloseProc;
    ProcessPtr : PPipeProcess
  end;

function CPipe (Path : CString; ArgV, Environment : PCStrings; PPipeStdIn, PPipeStdOut, PPipeStdErr : PCFilePtr) : Integer; asmname '_p_pipe';

function WaitPipeProcess (Process : PPipeProcess) : Boolean;
var
  WPID : Integer;
  Dummy : Boolean;
begin
  WaitPipeProcess := True;
  with Process^ do
    begin
      while Seconds <> 0 do
        begin
          if WaitPID (PID, Result^, Status^, False) = PID then Exit;
          Sleep (1);
          Dec (Seconds)
        end;
      if Signal <> 0 then Dummy := Kill (PID, Signal);
      if Wait then
        begin
          WPID := WaitPID (PID, Result^, Status^, True);
          if WPID <= 0 then WaitPipeProcess := False
        end
    end;
  Dispose (Process)
end;

procedure PipeTFDDClose (var PrivateData);
var
  FileProcess : TFileProcess absolute PrivateData;
  OpenFunc   : TOpenFunc;
  SelectProc : TSelectProc;
  ReadFunc   : TReadFunc;
  WriteFunc  : TWriteFunc;
  FlushProc  : TFlushProc;
  CloseProc  : TCloseProc;
  DoneProc   : TDoneProc;
  FPrivateData : Pointer;
  Dummy : Boolean;
begin
  with FileProcess do
    begin
      GetTFDD (FilePtr^, OpenFunc, SelectProc, ReadFunc, WriteFunc, FlushProc, CloseProc, DoneProc, FPrivateData);
      SetTFDD (FilePtr^, OpenFunc, SelectProc, ReadFunc, WriteFunc, FlushProc, OldCloseProc, DoneProc, FPrivateData);
      Close (FilePtr^);
      Dec (ProcessPtr^.OpenPipes);
      if ProcessPtr^.OpenPipes = 0 then Dummy := WaitPipeProcess (ProcessPtr)
    end;
  Dispose (@FileProcess)
end;

procedure SetCloseAction (var f : AnyFile; PProcess : PPipeProcess);
var
  p : PFileProcess;
  OpenFunc   : TOpenFunc;
  SelectProc : TSelectProc;
  ReadFunc   : TReadFunc;
  WriteFunc  : TWriteFunc;
  FlushProc  : TFlushProc;
  DoneProc   : TDoneProc;
  PrivateData : Pointer;
begin
  New (p);
  p^.FilePtr := @f;
  p^.ProcessPtr := PProcess;
  Inc (PProcess^.OpenPipes);
  GetTFDD (f, OpenFunc, SelectProc, ReadFunc, WriteFunc, FlushProc, p^.OldCloseProc, DoneProc, PrivateData);
  SetTFDD (f, OpenFunc, SelectProc, ReadFunc, WriteFunc, FlushProc, PipeTFDDClose, DoneProc, p)
end;

procedure Pipe (var ToInputFile, FromOutputFile, FromStdErrFile : AnyFile; (*@@fjf265 const*) ProcessName : String; protected var Parameters : TPStrings; Environment : PCStrings; var Process : PPipeProcess);
var
  ParameterCount, i : Cardinal;
  PipeStdOut, PipeStdErr, PipeStdIn : CFilePtr;
  PPipeStdOut, PPipeStdErr, PPipeStdIn : PCFilePtr;
  PID : Integer;
  PProcess : PPipeProcess;
begin
  if @Parameters = nil then ParameterCount := 0 else ParameterCount := Parameters.Count;
  var CParameters : array [0 .. ParameterCount + 1] of CString;
  CParameters [0] := ProcessName;
  for i := 1 to ParameterCount do CParameters [i] := Parameters [i]^;
  CParameters [ParameterCount + 1] := nil;
  if @ToInputFile = nil then PPipeStdIn := nil else PPipeStdIn := @PipeStdIn;
  if @FromOutputFile = nil then PPipeStdOut := nil else PPipeStdOut := @PipeStdOut;
  if @FromStdErrFile = nil then PPipeStdErr := nil else
    if @FromStdErrFile = @FromOutputFile then PPipeStdErr := @PipeStdOut else PPipeStdErr := @PipeStdErr;
  (*@@fjf302*)var pp:tstring=FSearchExecutable (ProcessName, GetEnv (PathEnvVar));
  if pp = ''
    then PID := - 1
    else PID := CPipe (pp, PCStrings (@CParameters), Environment, PPipeStdIn, PPipeStdOut, PPipeStdErr);
  if PID <= 0 then
    begin
      IOErrorCString (EPipe, ProcessName);
      Exit
    end;
  New (PProcess);
  PProcess^.PID := PID;
  PProcess^.Signal := DefaultPipeSignal;
  PProcess^.Seconds := DefaultPipeSeconds;
  PProcess^.Wait := DefaultPipeWait;
  PProcess^.OpenPipes := 0;
  PProcess^.Result := nil;
  PProcess^.Status := nil;
  if @Process <> nil then Process := PProcess;
  if @ToInputFile <> nil then
    begin
      AssignCFile (ToInputFile, PipeStdIn);
      Rewrite (ToInputFile);
      SetCloseAction (ToInputFile, PProcess)
    end;
  if @FromOutputFile <> nil then
    begin
      AssignCFile (FromOutputFile, PipeStdOut);
      Reset (FromOutputFile);
      SetCloseAction (FromOutputFile, PProcess)
    end;
  if (@FromStdErrFile <> nil) and (@FromStdErrFile <> @FromOutputFile) then
    begin
      AssignCFile (FromStdErrFile, PipeStdErr);
      Reset (FromStdErrFile);
      SetCloseAction (FromStdErrFile, PProcess)
    end
end;

{$else}

{ NOTE: This emulation code is quite a mess! Be warned if you want to
  understand it, and don't make any quick changes here unless you fully
  understand it. }

function CPipe (Path : CString; ArgV, Environment : PCStrings; NameStdIn, NameStdOut, NameStdErr : CString) : Integer; asmname '_p_pipe';

type
  PPipeData = ^TPipeData;
  TPipeData = record
    ProcName, Path : CString;
    ParameterCount : Cardinal;
    ArgV, EnvP : PCStrings;
    NameStdOut, NameStdErr, NameStdIn : TString;
    CNameStdOut, CNameStdErr, CNameStdIn : CString;
    PToInputFile, PFromOutputFile, PFromStdErrFile : ^AnyFile;
    InternalToInputFile : Text;
    PipeProcess : TPipeProcess
  end;

  PFileProcess = ^TFileProcess;
  TFileProcess = record
    FilePtr : PAnyFile;
    OldCloseProc : TCloseProc;
    PipeDataPtr : PPipeData
  end;

{ TPipeProcess.PID actually holds the exit status here }

function WaitPipeProcess (Process : PPipeProcess) : Boolean;
begin
  if Process^.Result <> nil then Process^.Result^ := PIDExited;
  if Process^.Status <> nil then Process^.Status^ := Process^.PID;
  WaitPipeProcess := Process^.PID >= 0
end;

procedure DoPipe (var PipeData : TPipeData);
var
  i : Cardinal;
  t : Text;
  Dummy : Boolean;
begin
  with PipeData do
    begin
      PipeProcess.PID := CPipe (Path, ArgV, EnvP, CNameStdIn, CNameStdOut, CNameStdErr);
      if PToInputFile <> nil then
        begin
          Assign (t, NameStdIn);
          Erase (t)
        end;
      if PipeProcess.PID < 0 then IOErrorCString (EPipe, ProcName);
      Dispose (ProcName);
      Dispose (Path);
      for i := 0 to ParameterCount do Dispose (ArgV^[i]);
      Dispose (ArgV);
      if PipeProcess.PID < 0 then Exit;
      Dummy := WaitPipeProcess (@PipeProcess);
      if PFromOutputFile <> nil then
        begin
          Reset (PFromOutputFile^, NameStdOut);
          Assign (t, NameStdOut);
          Erase (t)
        end;
      if (PFromStdErrFile <> nil) and (PFromStdErrFile <> PFromOutputFile) then
        begin
          Reset (PFromStdErrFile^, NameStdErr);
          Assign (t, NameStdErr);
          Erase (t)
        end
    end
end;

function PipeInTFDDWrite (var PrivateData; const Buffer; Size : SizeType) : SizeType;
var
  Data : TPipeData absolute PrivateData;
  CharBuffer : array [1 .. Size] of Char absolute Buffer;
begin
  with Data do
    Write (InternalToInputFile, CharBuffer);
  PipeInTFDDWrite := Size
end;

procedure PipeInTFDDClose (var PrivateData);
var Data : TPipeData absolute PrivateData;
begin
  with Data do
    begin
      Close (InternalToInputFile);
      if PFromOutputFile <> nil then Close (PFromOutputFile^);
      if (PFromStdErrFile <> nil) and (PFromStdErrFile <> PFromOutputFile) then Close (PFromStdErrFile^);
      Dec (Data.PipeProcess.OpenPipes);
      DoPipe (Data);
      Dispose (@Data)
    end
end;

procedure PipeOutTFDDClose (var PrivateData);
var
  FileProcess : TFileProcess absolute PrivateData;
  OpenFunc   : TOpenFunc;
  SelectProc : TSelectProc;
  ReadFunc   : TReadFunc;
  WriteFunc  : TWriteFunc;
  FlushProc  : TFlushProc;
  CloseProc  : TCloseProc;
  DoneProc   : TDoneProc;
  FPrivateData : Pointer;
  Dummy : Boolean;
begin
  with FileProcess do
    begin
      GetTFDD (FilePtr^, OpenFunc, SelectProc, ReadFunc, WriteFunc, FlushProc, CloseProc, DoneProc, FPrivateData);
      SetTFDD (FilePtr^, OpenFunc, SelectProc, ReadFunc, WriteFunc, FlushProc, OldCloseProc, DoneProc, FPrivateData);
      Close (FilePtr^);
      Dec (PipeDataPtr^.PipeProcess.OpenPipes);
      if PipeDataPtr^.PipeProcess.OpenPipes = 0 then
        begin
          Dummy := WaitPipeProcess (@PipeDataPtr^.PipeProcess);
          Dispose (PipeDataPtr)
        end
    end;
  Dispose (@FileProcess)
end;

procedure SetCloseAction (var f : AnyFile; PipeData : PPipeData);
var
  p : PFileProcess;
  OpenFunc   : TOpenFunc;
  SelectProc : SelectProc;
  ReadFunc   : TReadFunc;
  WriteFunc  : TWriteFunc;
  FlushProc  : TFlushProc;
  DoneProc   : TDoneProc;
  PrivateData : Pointer;
begin
  New (p);
  p^.FilePtr := @f;
  p^.PipeDataPtr := PipeData;
  GetTFDD (f, OpenFunc, SelectProc, ReadFunc, WriteFunc, FlushProc, p^.OldCloseProc, DoneProc, PrivateData);
  SetTFDD (f, OpenFunc, SelectProc, ReadFunc, WriteFunc, FlushProc, PipeOutTFDDClose, DoneProc, p)
end;

{ The "pipes" that come from the process are reset to nothing until the
  process is started. Use a TFDD, not an empty file or the null device,
  so that IOSelect will not select these files. }
procedure ResetNothing (var f : AnyFile);
begin
  AssignTFDD (f, (*@@fjf258*)TOpenFunc(nil), TSelectProc(nil), TReadFunc(nil), TWriteFunc(nil), TFlushProc(nil), TCloseProc(nil), TDoneProc(nil), nil);
  Reset (f)
end;

procedure Pipe (var ToInputFile, FromOutputFile, FromStdErrFile : AnyFile; (*@@fjf265 const*) ProcessName : String; protected var Parameters : TPStrings; Environment : PCStrings; var Process : PPipeProcess);
var
  i : Cardinal;
  PipeData : PPipeData;
begin
  (*@@fjf302*)var pp:tstring=FSearchExecutable (ProcessName, GetEnv (PathEnvVar));
  if pp = '' then
    begin
      IOErrorCString (EPipe, ProcessName);
      Exit
    end;
  New (PipeData);
  with PipeData^ do
    begin
      ProcName := NewCString (ProcessName);
      Path := NewCString (pp);
      if @Parameters = nil then ParameterCount := 0 else ParameterCount := Parameters.Count;
      GetMem (ArgV, (ParameterCount + 2) * SizeOf (CString));
      ArgV^[0] := NewCString (ProcessName);
      for i := 1 to ParameterCount do ArgV^[i] := NewCString (Parameters [i]^);
      ArgV^[ParameterCount + 1] := nil;
      EnvP := Environment;
      PipeProcess.PID := - 1;
      PipeProcess.Signal := DefaultPipeSignal;
      PipeProcess.Seconds := DefaultPipeSeconds;
      PipeProcess.Wait := DefaultPipeWait;
      PipeProcess.OpenPipes := 0;
      PipeProcess.Result := nil;
      PipeProcess.Status := nil;
      if @Process <> nil then Process := @PipeProcess;
      PToInputFile := @ToInputFile;
      PFromOutputFile := @FromOutputFile;
      PFromStdErrFile := @FromStdErrFile;
      if @FromOutputFile = nil
        then CNameStdOut := nil
        else
          begin
            Inc (PipeProcess.OpenPipes);
            NameStdOut := GetTempFileName;
            CNameStdOut := NameStdOut;
            if @ToInputFile <> nil then ResetNothing (FromOutputFile)
          end;
      if @FromStdErrFile = nil
        then CNameStdErr := nil
        else
          if @FromStdErrFile = @FromOutputFile
            then CNameStdErr := CNameStdOut
            else
              begin
                Inc (PipeProcess.OpenPipes);
                NameStdErr := GetTempFileName;
                CNameStdErr := NameStdErr;
                if @ToInputFile <> nil then ResetNothing (FromStdErrFile)
              end;
      if @ToInputFile = nil
        then
          begin
            CNameStdIn := nil;
            DoPipe (PipeData^);
            if @FromOutputFile <> nil then
              SetCloseAction (FromOutputFile, PipeData);
            if (@FromStdErrFile <> nil) and (@FromStdErrFile <> @FromOutputFile) then
              SetCloseAction (FromStdErrFile, PipeData)
          end
        else
          begin
            Inc (PipeProcess.OpenPipes);
            NameStdIn := GetTempFileName;
            CNameStdIn := NameStdIn;
            Rewrite (InternalToInputFile, NameStdIn);
            AssignTFDD (ToInputFile, (*@@fjf258*)TOpenFunc(nil), TSelectProc(nil), TReadFunc(nil), PipeInTFDDWrite, TFlushProc(nil), PipeInTFDDClose, TDoneProc(nil), PipeData);
            Rewrite (ToInputFile)
          end
    end
end;

{$endif}

begin
  PipeInit
end.
