/*
Support routine for pipe.pas

Copyright (C) 1998-99 Free Software Foundation, Inc.

Author: Frank Heckenbach <frank@pascal.gnu.de>

This library is free software; you can redistribute it and/or modify
it under the terms of the GNU Library General Public License as published by
the Free Software Foundation, version 2.

This library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
License for more details.

You should have received a copy of the GNU Library General Public License
along with this library; see the file COPYING.LIB.  If not, write to
the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
MA 02111-1307, USA.
*/

#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>

typedef unsigned char Boolean;
#define FALSE 0
#define TRUE 1

#ifndef MSDOS

int _p_pipe (char *path, char *argv [], char *envp [], FILE **pipe_stdin, FILE **pipe_stdout, FILE **pipe_stderr);
int _p_pipe (char *path, char *argv [], char *envp [], FILE **pipe_stdin, FILE **pipe_stdout, FILE **pipe_stderr)
{
  int fd_stdout [2], fd_stderr [2], fd_stdin [2];
  pid_t pid;
  if (access (path, X_OK)
      || (pipe_stdout && pipe (fd_stdout))
      || (pipe_stderr && pipe_stderr != pipe_stdout && pipe (fd_stderr))
      || (pipe_stdin && pipe (fd_stdin))
      || ((pid = fork ()) < 0))
    return - 1;
  if (pid)
    {
      if (pipe_stdout)
        {
          close (fd_stdout [1]);
          *pipe_stdout = fdopen (fd_stdout [0], "r");
        }
      if (pipe_stderr && pipe_stderr != pipe_stdout)
        {
          close (fd_stderr [1]);
          *pipe_stderr = fdopen (fd_stderr [0], "r");
        }
      if (pipe_stdin)
        {
          close (fd_stdin [0]);
          *pipe_stdin = fdopen (fd_stdin [1], "w");
        }
      return pid;
    }
  else
    {
      if (pipe_stdout)
        {
          dup2  (fd_stdout [1], 1);
          close (fd_stdout [0]);
          close (fd_stdout [1]);
        }
      if (pipe_stderr)
        {
          if (pipe_stderr != pipe_stdout)
            {
              dup2  (fd_stderr [1], 2);
              close (fd_stderr [0]);
              close (fd_stderr [1]);
            }
          else
            dup2 (1, 2);
        }
      if (pipe_stdin)
        {
          dup2  (fd_stdin [0], 0);
          close (fd_stdin [0]);
          close (fd_stdin [1]);
        }
      execve (path, argv, envp);
      _exit (127);
      return - 1;
    }
}

#else

#include <fcntl.h>
#include <process.h>

int _p_pipe (char *path, char *argv [], char *envp [], char *name_stdin, char *name_stdout, char *name_stderr);
int _p_pipe (char *path, char *argv [], char *envp [], char *name_stdin, char *name_stdout, char *name_stderr)
{
  int save_stdout, save_stderr, save_stdin, h, r;
  if (access (path, X_OK)) return - 1;
  if (name_stdin)
    {
      save_stdin = dup (0);
      h = open (name_stdin, O_RDONLY, 0777);
      dup2 (h, 0);
      close (h);
    }
  if (name_stdout)
    {
      save_stdout = dup (1);
      h = open (name_stdout, O_WRONLY | O_CREAT, 0777);
      dup2 (h, 1);
      close (h);
    }
  if (name_stderr)
    {
      save_stderr = dup (2);
      if (name_stderr != name_stdout)
        {
          h = open (name_stderr, O_WRONLY | O_CREAT, 0777);
          dup2 (h, 2);
          close (h);
        }
      else
        dup2 (1, 2);
    }
  r = spawnve (P_WAIT, path, argv, envp);
  if (name_stdin)
    {
      dup2 (save_stdin, 0);
      close (save_stdin);
    }
  if (name_stdout)
    {
      dup2 (save_stdout, 1);
      close (save_stdout);
    }
  if (name_stderr)
    {
      dup2 (save_stderr, 2);
      close (save_stderr);
    }
  return r;
}

#endif

Boolean _p_kill (int PID, int Signal);
Boolean _p_kill (int PID, int Signal)
{
  if (kill (PID, Signal) == 0)
    return TRUE;
  else
    return FALSE;
}

typedef enum { PIDNothing, PIDExited, PIDSignaled, PIDStopped, PIDUnknown } TWaitPIDResult;

#ifndef WIFEXITED
#define WIFEXITED(S) (((S) & 0xff) == 0)
#endif
#ifndef WEXITSTATUS
#define WEXITSTATUS(S) (((S) & 0xff00) >> 8)
#endif
#ifndef WIFSIGNALED
#define WIFSIGNALED(S) (((S) & 0xff) != 0 && ((S) & 0xff) != 0x7f)
#endif
#ifndef WTERMSIG
#define WTERMSIG(S) ((S) & 0x7f)
#endif
#ifndef WIFSTOPPED
#define WIFSTOPPED(S) (((S) & 0xff) == 0x7f)
#endif
#ifndef WSTOPSIG
#define WSTOPSIG(S) (((S) & 0xff00) >> 8)
#endif

int _p_waitpid (int PID, TWaitPIDResult *Result, int *Status, Boolean Block);
int _p_waitpid (int PID, TWaitPIDResult *Result, int *Status, Boolean Block)
{
  int wpid, wstatus;
  do
    wpid = waitpid (PID, &wstatus, Block ? 0 : WNOHANG);
  while
    #ifdef EINTR
    (wpid < 0 && errno == EINTR);
    #else
    (0);
    #endif
  if (wpid <= 0)
    {
      if (Result) *Result = PIDNothing;
      if (Status) *Status = 0;
    }
  else if (WIFEXITED (wstatus))
    {
      if (Result) *Result = PIDExited;
      if (Status) *Status = WEXITSTATUS (wstatus);
    }
  else if (WIFSIGNALED (wstatus))
    {
      if (Result) *Result = PIDSignaled;
      if (Status) *Status = WTERMSIG (wstatus);
    }
  else if (WIFSTOPPED (wstatus))
    {
      if (Result) *Result = PIDStopped;
      if (Status) *Status = WSTOPSIG (wstatus);
    }
  else
    {
      if (Result) *Result = PIDUnknown;
      if (Status) *Status = 0;
    }
  return wpid;
}

int _p_SIGHUP    = 0, _p_SIGINT    = 0, _p_SIGQUIT = 0, _p_SIGILL    = 0,
    _p_SIGABRT   = 0, _p_SIGFPE    = 0, _p_SIGKILL = 0, _p_SIGSEGV   = 0,
    _p_SIGPIPE   = 0, _p_SIGALRM   = 0, _p_SIGTERM = 0, _p_SIGUSR1   = 0,
    _p_SIGUSR2   = 0, _p_SIGCHLD   = 0, _p_SIGCONT = 0, _p_SIGSTOP   = 0,
    _p_SIGTSTP   = 0, _p_SIGTTIN   = 0, _p_SIGTTOU = 0, _p_SIGTRAP   = 0,
    _p_SIGIOT    = 0, _p_SIGEMT    = 0, _p_SIGBUS  = 0, _p_SIGSYS    = 0,
    _p_SIGSTKFLT = 0, _p_SIGURG    = 0, _p_SIGIO   = 0, _p_SIGPOLL   = 0,
    _p_SIGCLD    = 0, _p_SIGXCPU   = 0, _p_SIGXFSZ = 0, _p_SIGVTALRM = 0,
    _p_SIGPROF   = 0, _p_SIGPWR    = 0, _p_SIGINFO = 0, _p_SIGLOST   = 0,
    _p_SIGWINCH  = 0, _p_SIGUNUSED = 0;

void _p_pipeinit (void);
void _p_pipeinit ()
{
#ifdef SIGHUP
  _p_SIGHUP = SIGHUP;
#endif
#ifdef SIGINT
  _p_SIGINT = SIGINT;
#endif
#ifdef SIGQUIT
  _p_SIGQUIT = SIGQUIT;
#endif
#ifdef SIGILL
  _p_SIGILL = SIGILL;
#endif
#ifdef SIGABRT
  _p_SIGABRT = SIGABRT;
#endif
#ifdef SIGFPE
  _p_SIGFPE = SIGFPE;
#endif
#ifdef SIGKILL
  _p_SIGKILL = SIGKILL;
#endif
#ifdef SIGSEGV
  _p_SIGSEGV = SIGSEGV;
#endif
#ifdef SIGPIPE
  _p_SIGPIPE = SIGPIPE;
#endif
#ifdef SIGALRM
  _p_SIGALRM = SIGALRM;
#endif
#ifdef SIGTERM
  _p_SIGTERM = SIGTERM;
#endif
#ifdef SIGUSR1
  _p_SIGUSR1 = SIGUSR1;
#endif
#ifdef SIGUSR2
  _p_SIGUSR2 = SIGUSR2;
#endif
#ifdef SIGCHLD
  _p_SIGCHLD = SIGCHLD;
#endif
#ifdef SIGCONT
  _p_SIGCONT = SIGCONT;
#endif
#ifdef SIGSTOP
  _p_SIGSTOP = SIGSTOP;
#endif
#ifdef SIGTSTP
  _p_SIGTSTP = SIGTSTP;
#endif
#ifdef SIGTTIN
  _p_SIGTTIN = SIGTTIN;
#endif
#ifdef SIGTTOU
  _p_SIGTTOU = SIGTTOU;
#endif
#ifdef SIGTRAP
  _p_SIGTRAP = SIGTRAP;
#endif
#ifdef SIGIOT
  _p_SIGIOT = SIGIOT;
#endif
#ifdef SIGEMT
  _p_SIGEMT = SIGEMT;
#endif
#ifdef SIGBUS
  _p_SIGBUS = SIGBUS;
#endif
#ifdef SIGSYS
  _p_SIGSYS = SIGSYS;
#endif
#ifdef SIGSTKFLT
  _p_SIGSTKFLT = SIGSTKFLT;
#endif
#ifdef SIGURG
  _p_SIGURG = SIGURG;
#endif
#ifdef SIGIO
  _p_SIGIO = SIGIO;
#endif
#ifdef SIGPOLL
  _p_SIGPOLL = SIGPOLL;
#endif
#ifdef SIGCLD
  _p_SIGCLD = SIGCLD;
#endif
#ifdef SIGXCPU
  _p_SIGXCPU = SIGXCPU;
#endif
#ifdef SIGXFSZ
  _p_SIGXFSZ = SIGXFSZ;
#endif
#ifdef SIGVTALRM
  _p_SIGVTALRM = SIGVTALRM;
#endif
#ifdef SIGPROF
  _p_SIGPROF = SIGPROF;
#endif
#ifdef SIGPWR
  _p_SIGPWR = SIGPWR;
#endif
#ifdef SIGINFO
  _p_SIGINFO = SIGINFO;
#endif
#ifdef SIGLOST
  _p_SIGLOST = SIGLOST;
#endif
#ifdef SIGWINCH
  _p_SIGWINCH = SIGWINCH;
#endif
#ifdef SIGUNUSED
  _p_SIGUNUSED = SIGUNUSED;
#endif
}
